#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "phantom.h"

/*#define DEBUG 1*/

#define FIRSTSIZE 64

static int increase_tlist (struct _typelist **tlist);
static char *skiptolf (char *str);
static char *skipspace (char *str);
static int add_tlist (struct _typelist **tlist, char *str);
static char *getext (char *fname);


void destroy_chain (struct _chain **head)
{
	struct _chain *child, *parent;

	parent = *head;
	while (parent) {
		child = parent->next;
		free (parent);
		parent = child;
	};
	*head = NULL;
}

void destroy_mimetype (struct _typelist **tlist)
{
#ifdef DEBUG
	printf ("destroy tlist %p\n", *tlist);
#endif

	if (!*tlist)	return;

	destroy_chain (&((*tlist)->head));
	free (*tlist);
	*tlist = NULL;
}

static int increase_tlist (struct _typelist **tlist)
{
	int i = (*tlist)->max * 2;
	struct _typelist *newlist;

#ifdef DEBUG
	printf ("start increase_tlist to %d\n", i);
#endif

	newlist = malloc (sizeof(struct _typelist)
						+ (i - 1) * sizeof(struct _typeext));
	if (!newlist)	return FALSE;

	newlist->max = i;
	newlist->head = (*tlist)->head;
	for (i = 0; i < (*tlist)->current; i++) {
		newlist->list[i] = (*tlist)->list[i];
	}
	newlist->current = i;
	free (*tlist);
	*tlist = newlist;

#ifdef DEBUG
	printf ("done increase_tlist\n");
#endif

	return TRUE;
}

static char *skiptolf (char *str)
{
	while (*str && *str != '\n')	str++;
	*str = '\0';

	return str;
}

static char *skipspace (char *str)
{
#ifdef DEBUG
	printf ("skiping space\n");
#endif

	while (*str) {
		if (*str != ' ' && *str != '\t'
				 && *str != '\n')	break;
		str++;
	}

	return str;
}

static int add_tlist (struct _typelist **tlist, char *str)
{
	struct _chain *newtype, *parent;
	char *ext, *r;

#ifdef DEBUG
	printf ("start add_tlist\n");
#endif

	/* skip mimetype-name */
	for (r = str; *r > ' '; r++);

	/* spaces only ? */
	if (r == str)	return TRUE;

	/* got mimetype */
	newtype = malloc (sizeof (struct _chain)
				+ (int)(skiptolf (r) - str));
	if (!newtype)	return FALSE;

	newtype->next = NULL;
	strcpy (newtype->body, str);

	if (parent = (*tlist)->head) {
		while (parent->next)	parent = parent->next;
		parent->next = newtype;
	} else {
		(*tlist)->head = newtype;
	}

	r += newtype->body - str;
	*r++ = '\0';

#ifdef DEBUG
	printf ("got newtype %s : %p\n", newtype->body, newtype);
#endif

	while (*r) {
		r = skipspace (r);

		if (*r) {
			if ((*tlist)->current >= (*tlist)->max) {
				if (!increase_tlist (tlist))	return FALSE;
			}

			/* get extention */
			for (ext = r; *r && *r > ' '; r++);
			if (*r)	*r++ = '\0';

			(*tlist)->list[(*tlist)->current].name = ext;
			(*tlist)->list[(*tlist)->current++].mimetype = newtype->body;
#ifdef DEBUG
	printf ("regist type : %s -> %s\n", ext, newtype->body);
#endif
		}
	}

#ifdef DEBUG
	printf ("end add_tlist\n");
#endif

	return TRUE;
}


struct _typelist *read_mimetype ()
{
	FILE *fp;
	struct _typelist *t = NULL;
	char *ptr, buf[STR_LEN];

	t = malloc (sizeof(struct _typelist)
						+ (FIRSTSIZE - 1) * sizeof(struct _typeext));
	if (!t) {
		puts ("memory exhausted\n");
		return NULL;
	}
#ifdef DEBUG
	printf ("alloc mimetype %p\n", t);
#endif

	t->current = 0;
	t->max = FIRSTSIZE;
	t->head = NULL;

	ptr = getenvdir (buf, "HOME");
#ifdef DEBUG
	printf ("got home : %s\n", buf);
#endif

	strcpy (ptr, ".mime.types");

	fp = fopen (buf, "rt");
	if (!fp) {
		strcpy (ptr, "mimetype");
		fp = fopen (buf, "rt");
	}
	if (!fp) {
		puts (".mime.types cannot open\n");
		destroy_mimetype (&t);
		return NULL;
	}

#ifdef DEBUG
	printf ("opened mimetype %s\n", buf);
#endif

	while (fgets (buf, sizeof(buf), fp)) {
		if (buf[0] != '#' && buf[0] != '\n') {
			if (!add_tlist (&t, buf)) {
				destroy_mimetype (&t);
				break;
			}
		}
	}
#ifdef DEBUG
	printf ("closed mimetype\n");
#endif

	fclose (fp);

	realloc (t, sizeof(struct _typelist)
				+ (t->current - 1) * sizeof(struct _typeext));

	return t;
}

/* search extention name */
static char *getext (char *fname)
{
	char *p = fname, *lastperiod = NULL;

	while (*p) {
		if (*p == '.')	lastperiod = p;
		p++;
	}
	return lastperiod;
}


char *search_mimetype (struct _typelist *t, char *fname)
{
	int i;
	char *ptr = getext (fname);
	static char octet[] = "application/octet-stream";

	if (t && ptr) {
		ptr++;
		for (i = 0; i < t->current; i++) {
			if (!strcmpi (t->list[i].name, ptr)) {
				return t->list[i].mimetype;
			}
		}
	}
	return octet;
}

